# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
import time
from typing import Union, Dict, Any
import yaml
import random

import numpy as np
import cupy as cp
from tqdm.auto import tqdm

from ...electronics._camera_geometry import SiPMCameraGeometry
from ...electronics._camera import CherenkovSiPMCamera
from ...electronics._waveforms import Waveform
from ...electronics._pde import PhotonDetectionEfficiency
from ...electronics.signals.discriminator_signals._disciminator_kernels import (
    last_trigger_time_counter_from_camera_trigger,
    count_pixel_triggers
)
from ...electronics.signals.trigger_logic._trigger_kernels import (
    topological_camera_trigger,
    count_topological_camera_triggers
)
from ...electronics.signals.sampling_signals import digitize
from ...electronics.sources._poisson_flash import generate_poisson_flash

class AstriCameraGeometry(SiPMCameraGeometry):
    """ASTRI camera geometry class.

    See Also
    --------
        :py:class:`iactsim.electronics.SiPMCameraGeometry`: 
    """
    # Number of pdms per row
    _rows = np.array([3, 5, 7, 7, 7, 5, 3])

    # Maxium number of pdm in a row
    _max_per_row = np.max(_rows)

    #_pdms_below[i] is the number of pdms below the row i
    _pdms_below = np.append(np.array([0]), np.cumsum(_rows))

    # Total number of pdm
    _number_of_pdm = _pdms_below[-1]

    # Number of pixels per PDM side
    _n_pix_pdm = 8

    def __init__(self):
        super().__init__()

        # TODO: should it be used?
        self.position = np.array([0,0,0])
        """Position of the center of the central SiPM module (PDM 19) in the focal surface reference system."""

        self.pixels_separation = 0.225

        self.pixel_active_side = 6.95
    
    @property
    def body_dimensions(self):
        return np.asarray([0,0,0])
    
    @property
    def position(self):
        return self.__position
    
    @position.setter
    def position(self, p):
        self.__position = p
        self.__pdms_p, self.__pdms_n = self.pdms_position_and_rotation()
    
    @property
    def modules_p(self):
        return self.__pdms_p

    @property
    def modules_n(self):
        return self.__pdms_n

    @property
    def module_side(self):
        return self.pixel_active_side*self._n_pix_pdm + (self._n_pix_pdm-1)*self.pixels_separation

    @property
    def pixel_active_side(self):
        return self._pixel_active_side
    
    @pixel_active_side.setter
    def pixel_active_side(self, a_side):
        self._pixel_active_side = a_side

    @property
    def pixels_separation(self):
        return self._pixels_separation

    @pixels_separation.setter
    def pixels_separation(self, a_sep):
        self._pixels_separation = a_sep
    
    def pdmid_to_indices(self, k):
        # Out of range check
        if k<1 or k>self._number_of_pdm: 
            return None, None
        for j in range(len(self._rows)): 
            # Check if the pdm is in the row j
            if k <= self._pdms_below[j+1]:
                # Position of the pdm k starting from the first pdm of the row
                on_row_pos = k - self._pdms_below[j] - 1
                # Missing pdms at the beginning of the row
                n_missing = int((self._max_per_row - self._rows[j])/2)
                # Column i
                i = on_row_pos + n_missing
                return i, j

    def pdms_position_and_rotation(self):
        """Compute and return ASTRI photon detection modules position and normal vector with respect to the focal surface reference system."""
        positions = np.empty((37,3), dtype=np.float64)

        pdm_angular_sep = 3.1945 # degree
        for pdm in range(1,38):
            i, j = self.pdmid_to_indices(pdm)
            # signs and optical axis orientation:
            # (0,0,1) -> beta +, gamma -
            # (0,0,-1) -> beta -, gamma +
            beta = pdm_angular_sep * (i-7//2) * np.pi/180.
            gamma = - pdm_angular_sep * (j-7//2) * np.pi/180.

            sb = np.sin(beta)
            cb = np.cos(beta)
            sg = np.sin(gamma)
            cg = np.cos(gamma)

            # Rotate x then rotate y
            R = np.asarray(
                [
                    [    cb,  0,     sb],
                    [ sg*sb, cg, -sg*cb],
                    [-sb*cg, sg,  cb*cg]
                ]
            )
            
            # PDM position with origin on the focal plane sphere center
            positions[pdm-1] = R.dot([0,0,1060.])
        
        # Normal direction of the PDM surface
        directions = positions/np.linalg.norm(positions,axis=-1)[:,np.newaxis]
        
        # Move PDMs to the camera reference frame (origin at the center of PDM 19)
        positions[:,0] += self.position[0]
        positions[:,1] += self.position[1]
        positions[:,2] += self.position[2] - 1060.
        
        return positions, directions

class AstriCherenkovCamera(CherenkovSiPMCamera):
    """
    ASTRI Cherenkov camera class.

    See Also
    --------
        :py:class:`iactsim.electronics.CherenkovCamera`:
        :py:class:`iactsim.electronics.CherenkovSiPMCamera`:
    """
    _left_asic = np.asarray([0, 1, 2, 3, 8, 9, 10, 11, 16, 17, 18, 19, 24, 25, 26, 27, 32, 33, 34, 35, 40, 41, 42, 43, 48, 49, 50, 51, 56, 57, 58, 59])
    _right_asic = np.asarray([4, 5, 6, 7, 12, 13, 14, 15, 20, 21, 22, 23, 28, 29, 30, 31, 36, 37, 38, 39, 44, 45, 46, 47, 52, 53, 54, 55, 60, 61, 62, 63])

    def __init__(
            self,
        ):
        trigger_channels = [0]
        sampling_channels = [1,2]
        geometry = AstriCameraGeometry()
        waveforms = [Waveform()]*4
        channels_time_resolution = [1,1,1,1]
        pde = PhotonDetectionEfficiency()
        n_pixels = 2368

        # Initialize the base class
        super().__init__(
            n_pixels,
            waveforms,
            trigger_channels,
            sampling_channels,
            channels_time_resolution,
            pde,
            geometry,
        )        

        self.channels_gain[0] = cp.full((self.n_pixels,), 1., dtype=cp.float32)
        self.channels_gain[1] = cp.full((self.n_pixels,), 100, dtype=cp.float32)
        self.channels_gain[2] = cp.full((self.n_pixels,), 10, dtype=cp.float32)
        self.channels_gain[3] = self.channels_gain[0]

        # "Time" discriminator
        self.threshold_1 = cp.full((self.n_pixels,), 3, cp.float32)
        self.threshold_1_offset = cp.zeros((self.n_pixels,), dtype=cp.float32)
        self.n_contiguous = 5
        self.discriminator1_signals = None
        self.discriminator1_time_over_threshold = 1 # ns
        
        # Sampling info
        self.enable_peak_detection = True
        # ADCs baseline
        self.hg_adc_offset = cp.full((self.n_pixels,), 2**13, dtype=cp.float32)
        self.lg_adc_offset = cp.full((self.n_pixels,), 2**13, dtype=cp.float32)
        # ADCs noise
        self.hg_adc_noise = cp.full((self.n_pixels,), 10, dtype=cp.float32)
        self.lg_adc_noise = cp.full((self.n_pixels,), 2, dtype=cp.float32)
        self.adc_resolution = 14
        # Peak detection window extent
        self.sampling_window_extent = [45.,45.]
        # Time delay between a PDM triger and the peak-detector activation
        self.sampling_delay = [32., 32.]

        self.auxiliary_delay = [-128.]
        self.auxiliary_window_extent = [256.]

        self.peak_detection_window_extent = 32.

        self.sigma_ucells = np.full((self.n_pixels,), 0.01) 
        self.cross_talk = np.full((self.n_pixels,), 0.03)
        
        # Extend the trigger window by the trigger waveform peaking time
        self.trigger_window_end_offset = 30.

        # Calibration system
        self.blue_laser_illumination = cp.full((self.n_pixels,), 2, cp.float32)
        self.green_laser_illumination = cp.full((self.n_pixels,), 2, cp.float32)
        self.dark_rate = cp.full((self.n_pixels,), 0.001, cp.float32)

        # Seed
        self.increment_seed = True

        # Data
        self._peak_time_hg = cp.empty((self.n_pixels,), dtype=cp.float32)
        self._peak_time_lg = cp.empty((self.n_pixels,), dtype=cp.float32)
        self._peak_amplitude_hg = cp.empty((self.n_pixels,), dtype=cp.float32)
        self._peak_amplitude_lg = cp.empty((self.n_pixels,), dtype=cp.float32)
        self._trigger_time = []
        self._triggered_pdm = []
        self._pixel_trigger_time = []
        self._high_gain = []
        self._low_gain = []

        self.register_pixel_trigger_time = True

        self.show_progress = False

        self.restart()

    def restart(self):
        super().restart()
        self._event_counter = 0
        self._high_gain = []
        self._low_gain = []
        self._trigger_time = []
        self._triggered_pdm = []
        self._pixel_trigger_time = []
        self.timer.clear()

        left = self._left_asic
        right = self._right_asic
        self.pe_to_dac = cp.empty((37,64), dtype=cp.float32)
        peq = self.channels_gain[0].reshape(37,64)
        for k in range(37):
            k_peq = peq[k]
            self.pe_to_dac[k,left] = cp.nanmedian(k_peq[left])
            self.pe_to_dac[k,right] = cp.nanmedian(k_peq[right])

    @property
    def peak_detection_window_extent(self):
        return self._peak_detection_window_extent
    
    @peak_detection_window_extent.setter
    def peak_detection_window_extent(self, a_extent):
        # clock maxium delay
        dt = 1e3/200. + 1e3/300.
        self.sampling_window_extent[0] = a_extent + dt
        self.sampling_window_extent[1] = a_extent + dt
        self._peak_detection_window_extent = a_extent

    @property
    def threshold(self):
        return self._threshold

    @threshold.setter
    def threshold(self, pe):
        """Compute the disciminator thresholds for a given global threshold.
        That is, performs the so-called pe-to-DAC conversion.

        Parameters
        ----------
        pe : float
            Global threshold in photo-electrons.
        """
        self.threshold_1 = cp.round_(pe*self.pe_to_dac)
        self._threshold = pe

    @property
    def triggered_pdm(self):
        if len(self._triggered_pdm) == 0:
            return None
        else:
            return np.asarray(self._triggered_pdm)

    @property
    def high_gain(self):
        if len(self._high_gain) == 0:
            return None
        else:
            return np.stack([hg.get() for hg in self._high_gain])

    @property
    def low_gain(self):
        if len(self._low_gain) == 0:
            return None
        else:
            return np.stack([lg.get() for lg in self._low_gain])

    @property
    def pixel_trigger_time(self):
        if len(self._pixel_trigger_time) == 0:
            return None
        else:
            return np.stack([tt.get() for tt in self._pixel_trigger_time])

    @property
    def adc_resolution(self):
        return self._adc_resolution

    @adc_resolution.setter
    def adc_resolution(self, a_res):
        self._adc_max_value = cp.int32(2**a_res-1)
        self._adc_resolution = a_res
    
    @property
    def threshold_1(self):
        return self._threshold_1

    @threshold_1.setter
    def threshold_1(self, thrs):
        if not isinstance(thrs, cp.ndarray):
            thrs = cp.asarray(thrs, dtype=cp.float32)
        self._threshold_1 = thrs

    @property
    def hg_adc_noise(self):
        return self._hg_adc_noise

    @hg_adc_noise.setter
    def hg_adc_noise(self, noise):
        if not isinstance(noise, cp.ndarray):
            noise = cp.asarray(noise, dtype=cp.float32)
        self._hg_adc_noise = noise

    @property
    def hg_adc_offset(self):
        return self._hg_adc_offset

    @hg_adc_offset.setter
    def hg_adc_offset(self, offset):
        if not isinstance(offset, cp.ndarray):
            offset = cp.asarray(offset, dtype=cp.float32)
        self._hg_adc_offset = offset

    @property
    def lg_adc_noise(self):
        return self._lg_adc_noise

    @lg_adc_noise.setter
    def lg_adc_noise(self, noise):
        if not isinstance(noise, cp.ndarray):
            noise = cp.asarray(noise, dtype=cp.float32)
        self._lg_adc_noise = noise

    @property
    def lg_adc_offset(self):
        return self._lg_adc_offset

    @lg_adc_offset.setter
    def lg_adc_offset(self, offset):
        if not isinstance(offset, cp.ndarray):
            offset = cp.asarray(offset, dtype=cp.float32)
        self._lg_adc_offset = offset
        
    def trigger_action(self):
        """Generate a camera trigger.
        """
        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t0 = time.time()
            self.timer.add_section('trigger_action')
            
        self.discriminator1_signals = self.apply_ideal_discriminator_to_channel(
            0,
            self.threshold_1,
            self.threshold_1_offset,
            self.discriminator1_time_over_threshold/self.channels_time_resolution[0]
        )

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t1 = time.time()
            self.timer.add_entry('trigger_action', 'discriminator', t1-t0)
        
        n_pdms = len(self.geometry.modules_p)
        
        pdm_dimension = cp.int32(self.geometry.module_side/ self.geometry.pixel_active_side)
        window_size = cp.int32(len(self.time_windows[0]))
        n_contiguous = cp.int32(self.n_contiguous)
        self.pdms_trigger_signals = cp.empty((n_pdms,len(self.time_windows[0])), dtype=cp.float32)
        trigger_t_idices = cp.empty((n_pdms,), dtype=cp.int32)
        trigger = cp.empty((2,), cp.int32)

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t2 = time.time()
            self.timer.add_entry('trigger_action', 'copy to device', t2-t1)

        trigger_args = (
            self.discriminator1_signals,
            n_contiguous,
            self.pdms_trigger_signals,
            trigger_t_idices,
            trigger,
            window_size,
            pdm_dimension,
            cp.int32(n_pdms),
            self.seed[0]
        )
        topological_camera_trigger((n_pdms,), (128,), trigger_args)

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t3 = time.time()
            self.timer.add_entry('trigger_action', 'camera trigger', t3-t2)
        
        t_index, pdm_id = trigger.get()
        if t_index >= 0:
            self.triggered = True
            self.trigger_time = self.time_windows[0][t_index]
            self._triggered_pdm.append(pdm_id+1) # count from 1 to match the usual ID
            self._trigger_time.append(self.trigger_time)
        else:
            self.trigger_time = None

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t4 = time.time()
            self.timer.add_entry('trigger_action', 'copy to host', t4-t3)
    
    def sampling_action(self):
        """Perform peak detection and pixels trigger time measurement.
        """
        dt = np.random.uniform(0, 1e3/200 + 1e3/300)
        self._t0_peak_detection_hg = self.trigger_time + self.sampling_delay[0] + dt
        self._t0_peak_detection_lg = self.trigger_time + self.sampling_delay[1] + dt

        if self.enable_peak_detection:
            extent_hg = self.peak_detection_window_extent
            extent_lg = self.peak_detection_window_extent
            self.apply_peak_detection_to_channel(self._peak_amplitude_hg, self._peak_time_hg, 1, self._t0_peak_detection_hg, extent_hg)
            self.apply_peak_detection_to_channel(self._peak_amplitude_lg, self._peak_time_lg, 2, self._t0_peak_detection_lg, extent_lg)
        else:
            t_slice_hg = np.abs(self.time_windows[1] - self._t0_peak_detection_hg).argmin()
            t_slice_lg = np.abs(self.time_windows[2] - self._t0_peak_detection_lg).argmin()
            self._peak_amplitude_hg = self.get_channel_signals(1,reshape=True)[:,t_slice_hg].flatten()
            self._peak_amplitude_lg = self.get_channel_signals(2,reshape=True)[:,t_slice_lg].flatten()
        
        dig_hg = cp.empty((self.n_pixels,), dtype=cp.uint16)
        hg_args = (
            dig_hg,
            self._peak_amplitude_hg,
            self.hg_adc_offset,
            self.hg_adc_noise,
            self.seed,
            self._adc_max_value,
            cp.int32(self.n_pixels),
            cp.int32(1)
        )
        dig_lg = cp.empty((self.n_pixels,), dtype=cp.uint16)
        lg_args = (
            dig_lg,
            self._peak_amplitude_lg,
            self.lg_adc_offset,
            self.lg_adc_noise,
            self.seed,
            self._adc_max_value,
            cp.int32(self.n_pixels),
            cp.int32(1)
        )
        
        digitize((self.n_pixels,),(32,),hg_args)
        digitize((self.n_pixels,),(32,),lg_args)

        self._high_gain.append(dig_hg)
        self._low_gain.append(dig_lg)

        ## Trigger time
        if self.register_pixel_trigger_time:
            tt = cp.empty((self.n_pixels,), dtype=cp.int32)
            self.compute_signals([3])
            self._time_discriminator_signals = self.apply_ideal_discriminator_to_channel(
                3,
                self.threshold_1,
                self.threshold_1_offset,
                self.discriminator1_time_over_threshold/self.channels_time_resolution[3]
            )
            delay = 128. + self.auxiliary_delay[0]
            cam_trigger_clock = cp.int32((np.abs(self.time_windows[3] - (self.trigger_time+delay))).argmin())
            stop_clock = cp.int32(128)
            register_n_bits = cp.int32(8)
            wsize = len(self.time_windows[3])
            n_pixels = cp.int32(self.n_pixels)
            n_threads = self._number_of_threads(wsize)
            last_trigger_time_counter_from_camera_trigger(
                (self.n_pixels,),
                (n_threads,),
                (
                    tt,
                    self._time_discriminator_signals,
                    cam_trigger_clock,
                    stop_clock,
                    register_n_bits,
                    cp.int32(wsize), 
                    n_pixels
                ),
                shared_mem=4*(wsize+n_threads)
            )
            self._pixel_trigger_time.append(tt)

    def acquire_pedestal(self, n_frame):
        """Acquire pulse-height distribution with an external trigger and the current background.

        Parameters
        ----------
        n_frame : int
            number of frame to be acquired.
        """
        self.restart()

        # Save current configuration
        old_enable_camera_trigger = self.enable_camera_trigger
        old_fixed_windows = self.fixed_time_windows

        # Pedestal configuration
        self.trigger_time = 0
        self.enable_camera_trigger = False
        self.fixed_time_windows = True
        self.time_windows = [
            np.asarray([0],dtype=np.float32),
            np.arange(self.sampling_delay[0], self.sampling_delay[0]+self.sampling_window_extent[0], self.channels_time_resolution[1], dtype=np.float32),
            np.arange(self.sampling_delay[1], self.sampling_delay[1]+self.sampling_window_extent[1], self.channels_time_resolution[2], dtype=np.float32),
            np.arange(-128, 128., self.channels_time_resolution[3], dtype=np.float32),
        ]
        
        # Pedestal
        for i in tqdm(range(n_frame), disable=not self.show_progress):
            self.simulate_response(source=None)

        # Restore old configuration (not windows)
        self.enable_camera_trigger = old_enable_camera_trigger
        self.fixed_time_windows = old_fixed_windows
    
    def acquire_variance(self, n_sample):
        """Acquire variance data.

        Parameters
        ----------
        n_frame : int
            number of frame to be acquired.
        """
        self.restart()

        # Save current configuration
        old_fixed_windows = self.fixed_time_windows

        # Pedestal configuration
        self.trigger_time = 0
        self.fixed_time_windows = True
        self.time_windows = [
            np.asarray([0],dtype=np.float32),
            np.asarray([0],dtype=np.float32),
            np.asarray([0],dtype=np.float32),
            np.asarray([0],dtype=np.float32),
        ]
        
        # Pedestal
        self.source = None
        for _ in tqdm(range(n_sample), disable=not self.show_progress):
            self.compute_signals(self.sampling_channels)
            dig_hg = cp.empty((self.n_pixels,), dtype=cp.uint16)
            hg_args = (
                dig_hg,
                self.get_channel_signals(1),
                self.hg_adc_offset,
                self.hg_adc_noise,
                self.seed,
                self._adc_max_value,
                cp.int32(self.n_pixels),
                cp.int32(1)
            )
            dig_lg = cp.empty((self.n_pixels,), dtype=cp.uint16)
            lg_args = (
                dig_lg,
                self.get_channel_signals(2),
                self.lg_adc_offset,
                self.lg_adc_noise,
                self.seed,
                self._adc_max_value,
                cp.int32(self.n_pixels),
                cp.int32(1)
            )
            digitize((self.n_pixels,),(32,),hg_args)
            digitize((self.n_pixels,),(32,),lg_args)

            self._high_gain.append(dig_hg)
            self._low_gain.append(dig_lg)

            self.seed += self.n_pixels

        # Restore old configuration (not windows)
        self.fixed_time_windows = old_fixed_windows
    
    def acquire_dark_pedestal(self, n_frame):
        """Acquire pulse-height distribution with an external trigger and the closed lids background.

        Parameters
        ----------
        n_frame : int
            number of frame to be acquired.
        """
        old_background = self.background_rate.copy()
        self.background_rate = self.dark_rate

        self.acquire_pedestal(n_frame)

        self.background_rate = old_background

    def acquire_phd(self, n_frame):
        """Acquire pulse-height distribution with pulsed flash light.

        Parameters
        ----------
        n_frame : int
            number of frame to be acquired.
        """
        self.restart()
        
        # Save current configuration
        old_enable_camera_trigger = self.enable_camera_trigger
        old_fixed_windows = self.fixed_time_windows
        old_background = self.background_rate.copy()

        # PHD configuration
        self.background_rate = self.dark_rate
        self.enable_camera_trigger = True
        self.fixed_time_windows = True
        self.time_windows = [
            np.arange(0,  64, self.channels_time_resolution[0], np.float32),
            np.arange(0, 256, self.channels_time_resolution[1], np.float32),
            np.arange(0, 256, self.channels_time_resolution[2], np.float32),
            np.arange(0, 256, self.channels_time_resolution[3], np.float32),
        ]
        for i in tqdm(range(n_frame), disable=not self.show_progress):
            source = generate_poisson_flash(self.n_pixels, self.blue_laser_illumination, 0, 10, seed=None)
            self.simulate_response(source)

        # Restore previous configuration (but not time windows)
        self.enable_camera_trigger = old_enable_camera_trigger
        self.fixed_time_windows = old_fixed_windows
        self.background_rate = old_background
    
    def threshold_scan(self, start, stop, step, window_extent=512., max_duration=1e-3, maximum_count=None):
        """Measure the camera trigger rate as a function of the trigger threshold.

        Parameters
        ----------
        start : float
            Start threshold value.
        stop : float
            Stop threshold value.
        step : float
            Threshold step.
        window_extent : float, optional
            Extent in ns of the time window where to count triggers, by default 512 ns.
        max_duration : float, optional
            Integration time (in seconds) for each threshold, by default 1e-3 s.
        maximum_count : int, optional
            Stop when at least ``maximum_count`` triggers have been found at a each threshold, useful to speed-up the procedure at lower thresholds. By default None.

        Returns
        -------
        (numpy.ndarray,numpy.ndarray,numpy.ndarray,numpy.ndarray)
            Thresholds and corresponding rate, rate error and integration time. Integration times can be different for each threshold if a ``maximum_count`` is provided.
        """
        # Save current configuration
        old_fixed_windows = self.fixed_time_windows
        old_threshold = self.threshold
        
        self.restart()
        self.fixed_time_windows = True
        self.time_windows = [
            np.arange(0, window_extent, 10./3, np.float32),
            np.arange(0, 32, 1, np.float32),
            np.arange(0, 32, 1, np.float32),
            np.arange(0, 32, 1, np.float32),
        ]
        window_size = self.time_windows[0].shape[0]

        # Allow signals to be computed with no source
        self.source = None
        
        max_iterations = int(max_duration/window_extent*1e9)
        thresholds = np.arange(start, stop+step, step, dtype=np.float32)
        trigger_rate = cp.zeros(thresholds.shape, dtype=cp.int32)
        integration_time = np.zeros(thresholds.shape, dtype=np.float32)
        n_threads = self._number_of_threads(window_size)

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t_old = time.time()
            self.timer.add_section('trigger_scan')

        for i in tqdm(range(max_iterations), disable=not self.show_progress):
            self.compute_signals([0])
            for j in range(len(thresholds)):
                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_0 = time.time()

                # Skip if there are enough count at this threshold
                if maximum_count is not None:
                    if trigger_rate[j] > maximum_count:
                        continue
                
                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_1 = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}: check rate', t_th_1 - t_th_0)

                # Set the threshold
                self.threshold = thresholds[j]

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_2 = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}: set threshold', t_th_2 - t_th_1)

                # Get triggers
                self.trigger_action()

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_3 = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}: trigger action', t_th_3 - t_th_2)
                
                # Initialize count
                cam_trigger_counts = cp.zeros((1,), dtype=cp.int32)
                args = [
                    cam_trigger_counts,
                    self.pdms_trigger_signals,
                    cp.int32(window_size),
                    cp.int32(37)
                ]
                # Count triggers
                count_topological_camera_triggers((1,), (n_threads,), args, shared_mem=4*(window_size+n_threads))

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_4 = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}: count triggers', t_th_4 - t_th_3)

                integration_time[j] += window_extent / 1e9
                trigger_rate[j] += cam_trigger_counts[0]

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_5 = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}: update counts', t_th_5 - t_th_4)

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_new = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}', t_new-t_old)
                    t_old = t_new

            self.seed = self.seed + self.n_pixels
            self._event_counter += 1
        
        # Restore previous configuration (but not time windows)
        self.fixed_time_windows = old_fixed_windows
        self.threshold = old_threshold

        # Compute rate and erros
        integration_time = np.asarray(integration_time)
        trigger_rate = trigger_rate.get()
        error = np.sqrt(trigger_rate) / integration_time
        trigger_rate = trigger_rate / integration_time
        return thresholds, trigger_rate, error, integration_time

    def acquire_stairs(self, start, stop, step, integration_time=0.001):
        """Acquire staircase curves (pixel dark trigger rate as a function of threshold).

        Parameters
        ----------
        start : float
            Start threshold value.
        stop : float
            Stop threshold value.
        step : float
            Threshold step.
        integration_time : float, optional
            Integration time (in seconds), by default 0.001 s

        Returns
        -------
        (numpy.ndarray, numpy.ndarray)
            Thresholds and corresponding trigger rate per pixel.
        """
        old_background = self.background_rate.copy()

        self.background_rate = self.dark_rate

        self.timer.clear()
        self.restart()
        self.fixed_time_windows = True
        window_extent = 1024 #ns
        self.time_windows = [
            np.arange(0, window_extent, 10./3., np.float32),
            np.arange(0, 32, 1, np.float32),
            np.arange(0, 32, 1, np.float32),
        ]
        window_size = self.time_windows[0].shape[0]
        
        # Allow signals to be computed with no source
        self.source = None
        
        thrs = np.arange(start, stop+step, step)
        tr_counts = cp.zeros((self.n_pixels,len(thrs)), dtype=np.int32)
        n_iterations = int(integration_time*1e9/window_extent)
        for _ in tqdm(range(n_iterations), disable=not self.show_progress):
            self.compute_signals([0])
            for i,thr in enumerate(thrs):
                counts = cp.zeros((self.n_pixels,), dtype=cp.int32)
                threshold = cp.full((self.n_pixels,), thr, dtype=cp.float32)
                disc_signals = self.apply_ideal_discriminator_to_channel(
                    0,
                    threshold,
                    self.threshold_1_offset,
                    self.discriminator1_time_over_threshold/self.channels_time_resolution[0]
                )
                count_pixel_triggers((self.n_pixels,),(128,),(counts,disc_signals,cp.int32(window_size),cp.int32(self.n_pixels)),shared_mem=4*(window_size+128))
                tr_counts[:,i] += counts
            self.seed += self.n_pixels
        
        self.background_rate = old_background

        return thrs, tr_counts.get()

    def configure(self, config: Union[str, Path, Dict[str, Any]]) -> None:
        """Configure the camera with a yaml configuration file or a dictionary.

        Parameters
        ----------
        config : str, path-like, or dict
            Path to the configuration file (str or pathlib.Path),
            or a dictionary containing the configuration.
        """

        if isinstance(config, (str, Path)):
            with open(config, 'r') as file:
                cam_conf = yaml.safe_load(file)
        elif isinstance(config, dict):
            cam_conf = config
        else:
            raise TypeError(
                "config must be a string (path), pathlib.Path, or a dictionary, "
                f"not {type(config)}"
            )
        
        if 'waveforms' in cam_conf and cam_conf['waveforms'] is not None:
            waveforms = []
            for file in cam_conf['waveforms']:
                x, y = np.loadtxt(file).T
                wave = Waveform(x, y, 'AC')
                wave.normalize()
                waveforms.append(wave)
            self.waveforms = waveforms

        if 'pde' in cam_conf and cam_conf['pde'] is not None:
            data = np.loadtxt(cam_conf['pde']).T
            wl = data[0,:]
            pde = data[1,:]
            self.pde.value = pde
            self.pde.wavelength = wl

        if 'channels_gain' in cam_conf:
            for i,gain in enumerate(cam_conf['channels_gain']):
                if gain is not None:
                    if isinstance(gain, (int,float)):
                        array = cp.full((self.n_pixels,), gain, dtype=cp.float32)
                    else:
                        array = cp.loadtxt(gain, dtype=cp.float32)
                    if array.size != self.n_pixels:
                        raise(ValueError(f"Channel {i} gain: wrong number of values provided ({array.size}). Provide a single value or a value for each pixel ({self.n_pixels})."))
                    self.channels_gain[i] = array.flatten()

        if 'timer' in cam_conf and cam_conf['timer'] is not None:
            self.timer.active = cam_conf['timer']

        if 'seed' in cam_conf and cam_conf['seed'] is not None:
            seed = cam_conf['seed']
        else:
            seed = random.getrandbits(64)
        
        self.seed = cp.arange(0, self.n_pixels, 1, dtype=cp.uint64) + seed

        assign_directly = [
            'trigger_channels',
            'trigger_window_end_offset',
            'trigger_window_start_offset',
            'sampling_channels',
            'sampling_delay',
            'sampling_window_extent',
            'peak_detection_window_extent',
            'channels_time_resolution',
            'auxiliary_delay',
            'auxiliary_window_extent',
            'adc_resolution',
            'discriminator1_time_over_threshold',
            'register_pixel_trigger_time',
            'increment_seed',
            'show_progress',
            'threshold' # Must be assigned after after channels_gain[0]
        ]

        check_is_value = [
            'hg_adc_noise',
            'hg_adc_offset',
            'lg_adc_noise',
            'lg_adc_offset',
            'cross_talk',
            'sigma_ucells',
            'pixel_mask',
            'background_rate',
            'dark_rate',
            'blue_laser_illumination',
            'green_laser_illumination',
            'threshold_1_offset'
        ]

        for name in cam_conf.keys():
            value = cam_conf[name]
            if value is not None:
                if name in check_is_value:
                    dtype =cp.float32
                    if name == 'pixel_mask':
                        dtype = cp.int32
                    if isinstance(value, (int,float)):
                        array = cp.full((self.n_pixels,), value, dtype=dtype)
                    else:
                        array = cp.loadtxt(value, dtype=dtype)
                    if array.size != self.n_pixels:
                        raise(ValueError(f"{name}: wrong number of values provided ({array.size}). Provide a single value or a value for each pixel ({self.n_pixels})."))
                    setattr(self, name, array.flatten())
                if name in assign_directly:
                    setattr(self, name, value)
        
        self.restart()

    def plot_pixel_trigger_channel(self, pdm, pixel):
        import matplotlib.pyplot as plt

        pix = pixel - 1
        pdm = pdm - 1
        window = self.time_windows[0]
        signal = self.get_channel_signals(0, True)[64*pdm:64*(pdm+1)][pix].get()/self.channels_gain[0][pix].get()
        disc_signal = self.discriminator1_signals.reshape(37,64,-1)[pdm,pix].get()*self.threshold_1[pdm,pix].get()/self.channels_gain[0][pix].get()
        
        if self.triggered or not self.enable_camera_trigger and self.register_pixel_trigger_time:
            window = self.time_windows[3]
            signal = self.get_channel_signals(3, True)[64*pdm:64*(pdm+1)][pix].get()/self.channels_gain[3][pix].get()
            disc_signal = self._time_discriminator_signals.reshape(37,64,-1)[pdm,pix].get()*self.threshold_1[pdm,pix].get()/self.channels_gain[0][pix].get()
            plt.plot(window, signal, label='Fast signal')
            plt.vlines(self.trigger_time, signal.min(), signal.max(), color='black', ls='--', label='Camera Trigger')
            plt.plot(window, disc_signal, ls='-', lw=1, color='black', label='Pixel Trigger')
        else:
            plt.plot(window, signal, ls='-', lw=2, label='Fast signal')
            plt.plot(window, disc_signal, ls='-', lw=1, color='black', label='Pixel Trigger')

        plt.ylim(signal.min(), signal.max())
        
        plt.legend(title=f'PDM {pdm+1} Pixel {pix+1}')
        plt.grid(which='both')
    
    def plot_pdm_pixels_trigger_channel(self, pdm):
        from ...visualization._sipm_camera_plots import plot_sipm_module_pixels

        def plotf(pix):
            self.plot_pixel_trigger_channel(pdm, pix+1)
        
        plot_sipm_module_pixels(self, plotf, 5)

    def plot_pixel_sampling_channels(self, pdm, pixel):
        if self.enable_camera_trigger and not self.triggered:
            return

        import matplotlib.pyplot as plt

        pix = pixel - 1
        pdm = pdm - 1
        hg_window = self.time_windows[1]
        lg_window = self.time_windows[2]
        hg_signal = self.get_channel_signals(1, True)[64*pdm:64*(pdm+1)][pix].get()/self.channels_gain[1][pix].get()
        lg_signal = self.get_channel_signals(2, True)[64*pdm:64*(pdm+1)][pix].get()/self.channels_gain[2][pix].get()
        
        
        hg_plot = plt.plot(hg_window, hg_signal, label='HG signal')
        lg_plot = plt.plot(lg_window, lg_signal, label='LG signal')

        ymin = min(hg_signal.min(),lg_signal.min())
        ymax = max(hg_signal.max(),lg_signal.max())

        plt.ylim(ymin,ymax)
    
        if self.enable_peak_detection:
            t0 = self._t0_peak_detection_hg
            t1 = t0 + self.peak_detection_window_extent
            plt.fill_between(x=[t0,t1], y1=[ymin,ymin], y2=[ymax,ymax], color='grey', interpolate=True, alpha=0.2, label='PD interval')
            lg = self._peak_amplitude_lg[pdm*64+pix].get()/self.channels_gain[1][pix].get()
            hg = self._peak_amplitude_hg[pdm*64+pix].get()/self.channels_gain[2][pix].get()
            plt.hlines(hg, hg_window[0], hg_window[-1], color=hg_plot[0].get_color(), ls='--')
            plt.hlines(lg, lg_window[0], lg_window[-1], color=lg_plot[0].get_color(), ls='--')
            
        plt.legend(title=f'PDM {pdm+1} Pixel {pix+1}', framealpha=0.5)
        plt.grid(which='both')
    
    def plot_pdm_pixels_sampling_channels(self, pdm):
        from ...visualization._sipm_camera_plots import plot_sipm_module_pixels

        def plotf(pix):
            self.plot_pixel_sampling_channels(pdm, pix+1)
        
        plot_sipm_module_pixels(self, plotf, 5)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import time
from typing import (
    Tuple, List, 
    Union, Optional
)

import cupy as cp
import numpy as np
from tqdm.auto import tqdm

from .optics._optical_system import OpticalSystem
from .optics.utils._aspheric_coeffs import normalize_aspheric_coefficients
from .optics._cpu_transforms import local_to_telescope_rotation, photon_to_local_rotation
from .optics._surface_misc import SurfaceShape
from .optics._atmosphere import AtmosphericTransmission
from .optics._materials import Materials
from .optics import ray_tracing

from .electronics import CherenkovCamera, CherenkovSiPMCamera

from .utils._timer import BenchmarkTimer


class IACT:
    """A class to represent an Imaging Atmospheric Cherenkov Telescope (IACT).

    This class serves as a wrapper for a CUDA-accelerated backend. It handles the conversion 
    of optical system and camera information into CuPy ndarrays (transferring data to the GPU), 
    and arranges this data in a format suitable for input to the CUDA kernels that perform the 
    actual simulation.

    Parameters
    ----------
    optical_system : OpticalSystem
        An instance of a class derived from :py:class:`OpticalSystem` representing the 
        telescope optical system.
    camera : CherenkovCamera
        An instance of a class derived from :py:class:`CherenkovCamera` representing the 
        telescope Cherenkov camera.
    position : Tuple[float, float, float] | List[float] | numpy.ndarray
        The Cartesian coordinates (x, y, z) representing the telescope 
        position in millimeters. Can be a tuple, list, or NumPy array.
    pointing : Tuple[float, float] | List[float] | numpy.ndarray
        The horizontal coordinates (altitude, azimuth) representing the 
        telescope pointing direction in degrees. Can be a tuple, list, 
        or NumPy array.    
    
    Raises
    ------
    TypeError
        
        - If optical_system is not an instance of :py:class:`OpticalSystem` or a derived class.
        - If camera is not an instance of :py:class:`CherenkovCamera` or a derived class.

    ValueError
        
        - If position does not have 3 elements.
        - If pointing does not have 2 elements.

    """

    def __init__(
        self,
        optical_system: OpticalSystem,
        camera: CherenkovCamera = None,
        position: Tuple[float, float, float] | List[float] | np.ndarray = (0.,0.,0.),
        pointing: Tuple[float, float] | List[float] | np.ndarray = (0,0),
    ):
        if not isinstance(optical_system, OpticalSystem):
            raise TypeError("optical_system must be an instance of OpticalSystem or a derived class")

        # Validate and convert position to NumPy array
        if not isinstance(position, (tuple, list, np.ndarray)) or len(position) != 3:
            raise ValueError("position must be a tuple, list, or NumPy array of length 3")

        self._cuda_tracing_args = [None] * 25 # ray-tracing kernel related arguments
        self._cuda_atmosphere_args = [None] * 5 # atmosphere transmission related arguments
        self._cuda_materials_args = [None] * 5 # materials refractive index args
        
        self._position = np.asarray(position)

        self.pointing = pointing

        self.altitude = pointing[0]
        self.azimuth = pointing[1]

        self._optical_system = optical_system
        self._camera = camera

        self.atmospheric_transmission = AtmosphericTransmission()

        # Benchmark
        self.timer = BenchmarkTimer()
        self.timer.add_section('simulate_response')

        self.show_progress = False

    @property
    def position(self) -> np.ndarray:
        """
        numpy.ndarray: The telescope position in Cartesian coordinates (x, y, z) 
                    in millimeters.
        """
        return self._position

    @position.setter
    def position(self, value: Tuple[float, float, float] | List[float] | np.ndarray):
        if not isinstance(value, (tuple, list, np.ndarray)) or len(value) != 3:
            raise ValueError("position must be a tuple, list, or NumPy array of length 3")
        self._position = np.asarray(value)
        self._cuda_telescope_init()

    @property
    def pointing(self) -> np.ndarray:
        """
        numpy.ndarray: The telescope pointing direction in horizontal 
                    coordinates (altitude, azimuth) in degrees.
        """
        return self._pointing

    @pointing.setter
    def pointing(self, value: Tuple[float, float] | List[float] | np.ndarray):
        if not isinstance(value, (tuple, list, np.ndarray)) or len(value) != 2:
            raise ValueError("pointing must be a tuple, list, or NumPy array of length 2")
        self._pointing = np.asarray(value)
        self._altitude = value[0]
        self._azimuth = value[1]
        self._cuda_telescope_init()
    
    @property
    def altitude(self) -> float:
        """
        float: Telescope pointing altitude (0-90).
        """
        return self._altitude
    
    @altitude.setter
    def altitude(self, value: float):
        if value < 0 or value > 90:
            raise(ValueError('Telescope pointing altitude must be between 0 (horizon) and 90 (zenith) degrees.'))
        self._altitude = value
        self._pointing[0] = value
        self._cuda_telescope_init()

    @property
    def azimuth(self) -> float:
        """
        float: Telescope pointing azimuth (0-360).
        """
        return self._azimuth
    
    @azimuth.setter
    def azimuth(self, value: float):
        if value < 0 or value > 360:
            raise(ValueError('Telescope pointing azimuth must be between 0 and 360 degrees (increasing from north to east).'))
        self._altitude = value
        self._pointing[1] = value
        self._cuda_telescope_init()
    
    @property
    def optical_system(self) -> OpticalSystem:
        """
        OpticalSystem: The telescope optical system
        """
        return self._optical_system
    
    @property
    def camera(self) -> CherenkovCamera:
        """
        CherenkovCamera: The telescope Cherenkov camera.
        """
        return self._camera

    @camera.setter
    def camera(self, a_camera):
        if a_camera is not None and not isinstance(a_camera, CherenkovCamera):
            raise TypeError("camera must be an instance of CherenkovCamera or a derived class")
        self._camera = a_camera
    
    def _cuda_surfaces_shape_init(self):
        """Prepares optical surface data for CUDA-based ray tracing.

        This function extracts relevant properties from the optical surfaces
        defined in :py:attr:`optical_system` and converts them into CuPy arrays.
        """

        dtype = cp.float32

        n_surfaces = len(self.optical_system)

        n_asph_coef = 10
        
        # Allocate memory for surfaces info
        half_apertures = cp.zeros((n_surfaces*2,), dtype=dtype)   # Half-aperture for aspherical (rmin,rmax) surface and radius and height for cylindrical surfaces (r, height)
        aperture_shapes = cp.zeros((n_surfaces*2,), dtype=cp.int32)
        curvatures = cp.empty((n_surfaces,), dtype=dtype)
        conic_constants = cp.zeros((n_surfaces,), dtype=dtype)
        aspheric_coeffs = cp.zeros((n_surfaces, n_asph_coef), dtype=dtype) # pad with zeros
        positions = cp.empty(n_surfaces*3, dtype=dtype)
        offsets = cp.zeros(n_surfaces*2, dtype=dtype)
        rotations = cp.empty(n_surfaces*9, dtype=dtype)
        scatterings = cp.empty(n_surfaces, dtype=dtype)
        surface_flags = cp.zeros((n_surfaces*2,), dtype=cp.bool_)
        surface_types = cp.empty((n_surfaces,), dtype=cp.int32)
        surface_shapes = cp.empty((n_surfaces,), dtype=cp.int32)
        materials_in = cp.empty((n_surfaces,), dtype=cp.int32)
        materials_out = cp.empty((n_surfaces,), dtype=cp.int32)

        # Populate device arrays
        for i,surface in enumerate(self.optical_system):
            if surface._shape == SurfaceShape.ASPHERICAL:
                half_apertures[2*i] = surface.half_aperture
                half_apertures[2*i+1] = surface.central_hole_half_aperture
                curvatures[i] = surface.curvature
                aperture_shapes[2*i] = surface.aperture_shape.value
                aperture_shapes[2*i+1] = surface.central_hole_shape.value
                conic_constants[i] = surface.conic_constant
                surface_asph_coeffs = normalize_aspheric_coefficients(surface.aspheric_coefficients, surface.half_aperture)
                for j in range(len(surface_asph_coeffs)):
                    aspheric_coeffs[i][j] = surface_asph_coeffs[j]
                surface_flags[2*i] = surface.is_fresnel
                offsets[i*2] = surface.offset[0]
                offsets[i*2+1] = surface.offset[1]
            elif surface._shape == SurfaceShape.CYLINDRICAL:
                half_apertures[2*i] = surface.radius
                half_apertures[2*i+1] = surface.height
                surface_flags[2*i] = surface.top
                surface_flags[2*i+1] = surface.bottom
            else:
                raise(ValueError(f"Surface shape '{surface._shape}' ray-tracing not yet implemented."))
            
            positions[i*3] = surface.position[0]
            positions[i*3+1] = surface.position[1]
            positions[i*3+2] = surface.position[2]
            rotations[i*9:(i+1)*9] = cp.asarray(surface.get_rotation_matrix().flatten(), dtype=cp.float32)
            surface_types[i] = surface.type.value
            surface_shapes[i] = surface._shape.value
            scatterings[i] = surface.scattering_dispersion/180.
            materials_in[i] = surface.material_in.value
            materials_out[i] = surface.material_out.value

        # Refractive indices
        start_wls = []
        texture_pointers = []
        inv_dwls = []
        self._ri_texture_objects = []
        for material in Materials._members():
            obj, inv_dwl, start_wl = material.get_refractive_index_texture()
            self._ri_texture_objects.append(obj)
            texture_pointers.append(obj.ptr)
            inv_dwls.append(inv_dwl)
            start_wls.append(start_wl)
        start_wls = cp.asarray(start_wls, dtype=cp.float32)
        inv_dwls = cp.asarray(inv_dwls, dtype=cp.float32)
        tex_handlers_gpu = cp.asarray(texture_pointers, dtype=cp.uint64)

        self._cuda_tracing_args[:14] = [
            curvatures,
            conic_constants,
            aspheric_coeffs,
            surface_flags,
            half_apertures,
            aperture_shapes,
            positions,
            offsets,
            rotations,
            surface_types,
            surface_shapes,
            materials_in,
            materials_out,
            cp.int32(n_surfaces)
        ]

        self._cuda_tracing_args[16] = scatterings

        self._cuda_tracing_args[-3:] = [
            tex_handlers_gpu,
            inv_dwls,
            start_wls,
        ]

    def _cuda_telescope_init(self):
        # Position and rotation to transform into the telescope reference frame
        telescope_rot = cp.asarray(local_to_telescope_rotation(*self.pointing), dtype=cp.float32)
        telescope_pos = cp.asarray(self.position, dtype=cp.float32)
        
        self._cuda_tracing_args[14:16] = [
            telescope_rot,
            telescope_pos
        ]

    def _cuda_sipm_camera_geometry_init(self):
        # Number of modules
        n_pmds = len(self.camera.geometry.modules_n)
        # Modules rotation
        modules_r = np.empty((n_pmds*9,), dtype=np.float32)

        for i in range(n_pmds):
            modules_r[i*9:(i+1)*9] = photon_to_local_rotation(self.camera.geometry.modules_n[i]).flatten()

        self._trace_onto_sipm_modules_args = [ 
            cp.asarray(self.camera.geometry.modules_p, dtype=cp.float32),
            cp.asarray(modules_r, dtype=cp.float32),
            cp.float32(0.5*self.camera.geometry.module_side),
            cp.float32(self.camera.geometry.pixel_active_side),
            cp.float32(self.camera.geometry.pixels_separation),
            cp.int32(8),
            cp.int32(self.camera.geometry.modules_p.shape[0]),
        ]

        # Photon detection efficiency
        pde_curve = self.camera.pde
        wl = pde_curve.wavelength
        value = pde_curve.value
        if pde_curve is not None and wl is not None and value is not None:
            if np.any(value>1) or np.any(value<0):
                raise(ValueError('Photon detection efficiency must be between 0 and 1.'))
            if len(wl) != len(value):
                raise(ValueError('Photon detection efficiency value array and wavelength array must have the same number of elements.'))
            self._trace_onto_sipm_modules_args += [
                cp.asarray(wl, dtype=cp.float32),
                cp.asarray(value, dtype=cp.float32),
                cp.int32(len(wl)),
                cp.float32(1./(wl[1]-wl[0])),
            ]
        else:
            self._trace_onto_sipm_modules_args += [
                cp.empty((0,), dtype=cp.float32),
                cp.empty((0,), dtype=cp.float32),
                cp.int32(0),
                cp.float32(0.),
            ]
        
        # TODO: for now assume focal plane is surrounded only by AIR
        # find a way to easly expose this to the user
        obj, inv_dwl, start_wl = Materials.AIR.get_refractive_index_texture()
        self._trace_onto_sipm_modules_args += [
            cp.uint64(obj.ptr),
            cp.float32(inv_dwl),
            cp.float32(start_wl)
        ]

    def __check_efficiency_matrix(self, eff):
        if eff.value is None:
            return False
        
        if eff.wavelength is None and eff.incidence_angle is None:
            return False
        
        # Is ok since np.size(None) returns 1
        size = np.size(eff.wavelength)*np.size(eff.incidence_angle)

        if size == np.size(eff.value):
            return True

        return False

    def _cuda_surfaces_efficiency_init(self):
        # Transimission curve of each surface
        void = np.asarray([], dtype=np.float32)
        tr_curves = []
        tr_x1 = []
        tr_x2 = []
        for s in self.optical_system:
            eff = s.efficiency
            skip = not self.__check_efficiency_matrix(eff)
            if skip:
                tr_curves.append(void)
                tr_x1.append(void)
                tr_x2.append(void)
                continue
            tr_curves.append(eff.value)
            x1 = void if eff.wavelength is None else eff.wavelength
            x2 = void if eff.incidence_angle is None else eff.incidence_angle
            tr_x1.append(x1) 
            tr_x2.append(x2)

        # Start position inside arrays
        start_tr = cp.asarray(np.cumsum([0]+[tr.size for tr in tr_curves[:-1]]), dtype=cp.int32)
        start_x1 = cp.asarray(np.cumsum([0]+[len(x) for x in tr_x1[:-1]]), dtype=cp.int32)
        start_x2 = cp.asarray(np.cumsum([0]+[len(x) for x in tr_x2[:-1]]), dtype=cp.int32)
        # Sizes
        ns = cp.asarray([len(x) for x in tr_x1], cp.int32)
        ms = cp.asarray([len(x) for x in tr_x2], cp.int32)
        # All together
        tr_curve_sizes = cp.asarray(np.column_stack([start_tr, start_x1, start_x2, ns, ms]), dtype=cp.int32)

        # Flatten transmission values and move to GPU
        tr_curves = cp.asarray(np.concatenate([tr.flatten() for tr in tr_curves]), dtype=cp.float32)

        # Move x1 and x2 to GPU
        x1s = cp.asarray(np.concatenate(tr_x1), dtype=cp.float32)
        x2s = cp.asarray(np.concatenate(tr_x2), dtype=cp.float32)

        # Wavelength and angle steps
        inv_steps_x1 = cp.asarray([0 if len(x)<2 else 1./(x[1]-x[0]) for x in tr_x1], dtype=cp.float32)
        inv_steps_x2 = cp.asarray([0 if len(x)<2 else 1./(x[1]-x[0]) for x in tr_x2], dtype=cp.float32)
        inv_steps = cp.column_stack([inv_steps_x1, inv_steps_x2])

        self._cuda_tracing_args[17:22] = [
            tr_curves,
            x1s,
            x2s,
            tr_curve_sizes,
            inv_steps
        ]

    def _cuda_atmosphere_init(self):
        if self.atmospheric_transmission.value is not None:
            wl = self.atmospheric_transmission.wavelength
            zem = self.atmospheric_transmission.altitude
            self._cuda_atmosphere_args[:] = [
                cp.asarray(self.atmospheric_transmission.value, dtype=cp.float32),
                cp.asarray(wl, dtype=cp.float32),
                cp.asarray(zem, dtype=cp.float32),
                cp.asarray([len(wl), len(zem)], dtype=cp.int32),
            ]
                
    def cuda_init(self):
        """Build and copy to device all the CUDA kernels input related to the optical configuration."""

        # Ray-tracing throught the telescope optical system
        self._cuda_surfaces_shape_init()

        self._cuda_telescope_init()

        self._cuda_surfaces_efficiency_init()

        if isinstance(self.camera, CherenkovSiPMCamera):
            self._cuda_sipm_camera_geometry_init()

        # # Atmospheric transmission
        self._cuda_atmosphere_init()

    # TODO: revise _trace_onto_sipm_modules_args:
    #   - split it into sub-arg lists
    #   - init each-sub arg at run-time when a parameter is modified
    def trace_photons(self,
            positions: Union[np.ndarray, cp.ndarray],
            directions: Union[np.ndarray, cp.ndarray],
            wavelengths: Union[np.ndarray, cp.ndarray],
            arrival_times: Union[np.ndarray, cp.ndarray],
            emission_altitudes: Optional[Union[np.ndarray, cp.ndarray]] = None,
            events_mapping: Union[np.ndarray, list] = None,
            *,
            photons_per_bunch: int = 1,
            simulate_camera = False,
            min_n_pes = 1,
            trace_onto_camera: bool = True,
            get_camera_input: bool = False,
            
        ):
        """Traces the paths of photons through the telescope optical system and optionally simulates the Cherenkov camera response.

        This function simulates the propagation of photons through an optical system,
        taking into account their initial positions, directions, wavelengths, and arrival times.
        It can optionally:
        
          - model the effects of atmospheric transmission based on emission altitudes;
          - simulate the Cherenkov camera response.

        Each element in the input arrays can represent either a single photon or a "bunch" of photons.
        The ``photons_per_bunch`` parameter determines how many photons are represented by each element.
        If ``photons_per_bunch`` > 1, it is assumed that all photons within a bunch share the same
        initial position, direction, wavelength, and arrival time.

        Parameters
        ----------
        positions : Union[numpy.ndarray, cp.ndarray] of shape (N, 3)
            Initial Cartesian coordinates (x, y, z) of the photon bunches in millimeters (mm).
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        directions : Union[numpy.ndarray, cp.ndarray] of shape (N, 3)
            Initial direction vectors (vx, vy, vz) of the photon bunches.
            These should be normalized. Must be either a NumPy (CPU) or CuPy (GPU) array.
        wavelengths : Union[numpy.ndarray, cp.ndarray] of shape (N,)
            Wavelengths of the photon bunches in nanometers (nm).
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        arrival_times : Union[numpy.ndarray, cp.ndarray] of shape (N,)
            Arrival times of the photon bunches at their respective positions in nanoseconds (ns).
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        emission_altitudes : Optional[Union[numpy.ndarray, cp.ndarray]], optional
            Emission altitudes of the photon bunches in millimeters (mm).S
            This is used to calculate atmospheric transmission effects if enabled.
            Must be either a NumPy (CPU) or CuPy (GPU) array.
        events_mapping : numpy.ndarray or list
            A NumPy array of shape (N+1,) to split the photons in to groups corresponding to N different events when 
            generating the camera input arrays.
            ``events_mapping[i]`` indicates the starting index in ``positions`` for the i-th event.
            ``events_mapping[i+1]`` indicates the ending index (exclusive) in ``positions`` for the i-th event.
            Therefore, photons belonging to event i are located in ``positions[events_mapping[i]:events_mapping[i+1], :]``.
            The last element of events_mapping must be equal to the number of photons.
        photons_per_bunch : int, optional
            Number of photons represented by each element in the input arrays. If 1, each element represents a single photon.
            If > 1, each element represents a bunch of photons that share the same properties, by default 1.
        simulate_camera : bool, optional
            Whether to simulate the response of the telescope Cherenkov camera.
        min_n_pes : int, optional
            Minimum number of photo-electrons needed to trigger a camera simulation (if ``simulate_camera`` is True). By default 1.
        trace_onto_camera : bool, optional
            Whether to trace photons into camera pixels (if a camera is definined). It is set to True if ``simulate_camera`` is True.
            By default True.
        get_camera_input : bool, optional
            Wheter to return the input for camera simulation (i.e. photon arrival times for each pixel) as CuPy ndarrays.
            By default False.

        Returns
        -------
        ndarray
            List of triggered events index. If ``simulate_camera`` is True and ``get_camera_input`` is False.
        tuple(cp.ndarray, cp.ndaray)
            Input CuPy arrays for :py:meth:`CherenkovCamera.simulate_response()`` method. If ``simulate_camera`` is False and ``get_camera_input`` is True.
        tuple(cp.ndarray, cp.ndaray, np.ndarray)
            Input CuPy arrays for :py:meth:`CherenkovCamera.simulate_response()`` method and the list of triggered events index. If ``simulate_camera`` is True and ``get_camera_input`` is True.

        Warning
        -------
            The input arrays can be either CuPy or NumPy ndarray. In the former case arrays are modified in place.
            In the latter case the arrays are copied to the device and can be accessed through the following attributes:

              - self._d_directions
              - self._d_positions
              - self._d_wavelengths
              - self._d_arrival_times
              - self._d_emission_altitudes
              - self._d_weights
        
        Warning
        -------
            Telescope configuration must be copied into the device before calling this method. Use the :py:meth:`cuda_init()` method.

        Notes
        -----

          - The ray-tracing operates with positions in mm, wavelengths in nm, and times in ns. It also handles photon bunches,
            where each thread may process multiple photons if ``photons_per_bunch`` > 1. In this case is assumed 
            that each element of the input arrays represents a bunch of photons that share the same 
            initial position, direction, wavelength, and arrival time. The arrays will maintain this
            structure, with each element representing the updated properties of a photon bunch, but the bunch size may
            be reduced throught the ray-tracing due to the interacrtion with optical elements, atmosphere and pixels.

        """
        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t0 = time.time()

        blocksize = 256  # Threads per block
        n_photons = positions.shape[0] # Number of input elements
        num_blocks = int(np.ceil(n_photons / blocksize))

        self._d_arrival_times = cp.asarray(arrival_times, dtype=cp.float32)
        self._d_directions = cp.asarray(directions, dtype=cp.float32)
        self._d_positions = cp.asarray(positions, dtype=cp.float32)
        self._d_wavelengths = cp.asarray(wavelengths, dtype=cp.float32)

        if emission_altitudes is not None:
            self._d_emission_altitudes = cp.asarray(emission_altitudes, dtype=cp.float32)

        # Initialize photon bunches size on device
        # Each input is `photons_per_bunch` photons
        self._d_weights = cp.full((n_photons,), photons_per_bunch, dtype=cp.int32) 

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t1 = time.time()
            self.timer.add_entry('simulate_response', 'copy to device', t1-t0)
        
        # Optional atmospheric transmission calculation
        if self.atmospheric_transmission.value is not None and emission_altitudes is not None:
            ray_tracing.atmospheric_transmission(
                (num_blocks,), 
                (blocksize,),
                (
                    self._d_positions,
                    self._d_directions,
                    self._d_wavelengths,
                    self._d_arrival_times,
                    self._d_weights, 
                    self._d_emission_altitudes,
                    *self._cuda_atmosphere_args,
                    cp.int32(n_photons)
                )
            )

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t2 = time.time()
            self.timer.add_entry('simulate_response', 'atmospheric transmission', t2-t1)

        # Perform ray tracing on GPU
        ray_tracing.ray_tracing(
            (num_blocks,),
            (blocksize,),
            (
                self._d_positions,
                self._d_directions,
                self._d_wavelengths,
                self._d_arrival_times,
                self._d_weights,
                *self._cuda_tracing_args,
                cp.int32(n_photons)
            )
        )

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t3 = time.time()
            self.timer.add_entry('simulate_response', 'trace onto focal plane', t3-t2)

        # Trace onto SiPM camera
        if (trace_onto_camera or simulate_camera) and issubclass(type(self.camera), CherenkovSiPMCamera):
            # The id of the pixel the photon reaches
            self._d_pixid = cp.empty((n_photons,), dtype=cp.int32)
            ray_tracing.trace_onto_sipm_modules(
                (num_blocks,),
                (blocksize,),
                (
                    self._d_positions,
                    self._d_directions,
                    self._d_wavelengths,
                    self._d_arrival_times,
                    self._d_weights,
                    self._d_pixid,
                    cp.int32(n_photons),
                    *self._trace_onto_sipm_modules_args
                )
            )

            if self.timer.active:
                cp.cuda.stream.get_current_stream().synchronize()
                t4 = time.time()
                self.timer.add_entry('simulate_response', 'trace onto SiPM modules', t4-t3)

            # Build the input for `CherenkovCamera.simulate_response()` method
            if get_camera_input or simulate_camera:
                if events_mapping is None:
                    events_mapping = cp.asarray([0,n_photons], dtype=cp.int32)
                else:
                    events_mapping = cp.asarray(events_mapping, dtype=cp.int32)
                                
                n_events = len(events_mapping) - 1
                n_pixels = self.camera.n_pixels
                
                blocksize = 128
                nblock = int(np.ceil(n_events/blocksize))

                pix_counter = cp.zeros((n_events,n_pixels+1), dtype=cp.int32)
                counter = cp.zeros((n_photons,), dtype=cp.int32)
                gpu_ntot = cp.zeros((n_events+1,), dtype=cp.int32)
                ray_tracing.count_all_photons(
                    (nblock,),
                    (blocksize,),
                    (
                        self._d_weights,
                        self._d_pixid,
                        counter,
                        pix_counter,
                        events_mapping,
                        gpu_ntot,
                        cp.int32(n_pixels),
                        cp.int32(n_events)
                    )
                )

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t5 = time.time()
                    self.timer.add_entry('simulate_response', 'count pixel pes', t5-t4)
                
                ntot = cp.int32(np.sum(gpu_ntot.get(), dtype=np.int32))
                phs = cp.empty((ntot,), dtype=cp.float32)
                phs_mapping = cp.cumsum(pix_counter, dtype=cp.int32, axis=1)
                pe_mapping = cp.cumsum(gpu_ntot, dtype=cp.int32)
                ray_tracing.camera_inputs(
                    (nblock,),
                    (blocksize,),
                    (
                        phs,
                        phs_mapping,
                        self._d_arrival_times,
                        self._d_weights,
                        self._d_pixid,
                        counter,
                        pe_mapping,
                        events_mapping,
                        cp.int32(n_pixels),
                        cp.int32(n_events)
                    )
                )

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t6 = time.time()
                    self.timer.add_entry('simulate_response', 'generate camera input', t6-t5)

                if simulate_camera:
                    # Mapping onto the photo-electrons arrival time array (phs_mapping)
                    # Since phs_mapping contains the mapping for each individual event:
                    #    phs[event] -> phs_mapping[event]
                    # We need to add an offset to phs_mapping to take into account
                    # all the photons in the previous events in order to select a slice
                    # of phs that corresponds to a single event
                    phs_mapping_global = np.insert(np.cumsum(np.sum(np.diff(phs_mapping.get(),axis=1), axis=1)), 0, 0)
                    event_number = []
                    for i in tqdm(range(len(phs_mapping)), disable=not self.show_progress):
                        start_phs = phs_mapping_global[i]
                        end_phs = phs_mapping_global[i+1]
                        if end_phs-start_phs < min_n_pes:
                            continue
                        self.camera.simulate_response((phs[start_phs:end_phs], phs_mapping[i]))
                        if self.camera.triggered:
                            event_number.append(i)
                    
                    event_number = np.asarray(event_number)

                    if self.timer.active:
                        cp.cuda.stream.get_current_stream().synchronize()
                        t7 = time.time()
                        self.timer.add_entry('simulate_response', 'simulate camera response', t7-t6)

                    if get_camera_input:
                        return phs, phs_mapping, event_number
                    else:
                        return event_number
                
                if get_camera_input:
                    return phs, phs_mapping


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
from enum import Enum
import numpy as np

from ..visualization._iactsim_style import iactsim_style, MplInteractiveContext

class SurfaceType(Enum):
    """
    Enumeration representing the different types of optical surfaces.

    The directionality of reflection and sensitivity (e.g., "from above" or 
    "from below") is defined within the *local reference frame* of the surface at the point 
    of photon incidence. As a reference:

        - the front side of a surface with negative curvature is the convex side.
        - the front side of a surface with posistive curvature is the concave side.

    Attributes
    ----------
    REFLECTIVE : int
        Represents a reflective surface (both sides).
    REFLECTIVE_IN : int
        Represents a surface that is reflective on the front side. 
        In the surface's local reference frame, only photons arriving with a negative 
        z-component of their direction vector (i.e., coming "from the front", or 
        "from above" in the context of the local frame) are reflected. 
        The curvature of the surface does not affect this behavior.
    REFLECTIVE_OUT : int
        Represents a surface that is reflective on the back side. In the surface's 
        local reference frame, only photons arriving with a positive z-component of 
        their direction vector (i.e., coming "from the back", or "from below" in the 
        context of the local frame) are reflected. The curvature of the surface 
        does not affect this behavior.
    REFRACTIVE : int
        Represents a refractive surface. The refraction is the same on both sides.
    SENSITIVE : int
        Represents the focal plane surface where photons are detected. The sensitivity
        is the same on both sides.
    SENSITIVE_IN : int
        Represents a surface that is sensitive on the front side (as defined by the 
        local reference frame's positive z-axis). In the surface's local reference 
        frame, only photons arriving with a negative z-component of their direction 
        vector (i.e., coming "from the front", or "from above" in the context of the 
        local frame) are detected. The curvature of the surface does not affect 
        this behavior.
    SENSITIVE_OUT : int
        Represents a surface that is sensitive on the back side. In the surface's 
        local reference frame, only photons arriving with a positive z-component of 
        their direction vector (i.e., coming "from the back", or "from below" in the 
        context of the local frame) are detected. The curvature of the surface 
        does not affect this behavior.
    OPAQUE : int
        Represents a surface that blocks light.
    DUMMY : int
        Represents a surface that neither reflects nor refracts light. 
        It can be used to introduce artificial absorption or scattering effects, 
        serving as a means to model specific behaviors within the optical system.
    
    """
    REFLECTIVE = 0
    REFLECTIVE_IN = 1
    REFLECTIVE_OUT = 2
    REFRACTIVE = 3
    SENSITIVE = 4
    SENSITIVE_IN = 5
    SENSITIVE_OUT = 6
    OPAQUE = 7
    DUMMY = 8
    TEST_SENSITIVE = 9

class SurfaceShape(Enum):
    """
    Enumeration representing different types of surface shapes.

    This enum defines common surface shapes encountered in optical systems or
    other fields requiring precise geometrical definitions.

    """
    ASPHERICAL = 0
    """Represents an aspherical surface."""
    CYLINDRICAL = 1
    """Represents a cylindrical surface."""
    FLAT = 2
    """Represents a flat surface."""
    SPHERICAL = 3
    """Represents a spherical surface."""

class ApertureShape(Enum):
    """
    Enumeration representing the possible shapes of an aperture.

    Attributes
    ----------
    CIRCULAR : int
        Represents a circular aperture.
    HEXAGONAL : int
        Represents a hexagonal aperture.
    SQUARE : int
        Represents a square aperture.
    """
    CIRCULAR = 0
    HEXAGONAL = 1
    SQUARE = 2

@dataclass
class SurfaceEfficiency:
    """
    Represents surface transmittance/reflectance as a function of photon wavelength and incidence angle.
    """
    value: np.ndarray = field(default=None, metadata={'units': 'None'})
    """A numpy.ndarray of shape (n, m) or (n*m,) representing the transmittance/reflectance value (from 0 to 1).
    The value at wavelength[j] and incidence_angle[i] must correspond to:

        - value[i*n+j], if the array is flattened;
        - value[i,j], otherwise.
    
    """

    wavelength: np.ndarray = field(default=None, metadata={'units': 'nanometers'})
    """A NumPy ndarray of shape (n,) representing the wavelengths in nanometers."""

    incidence_angle: np.ndarray = field(default=None, metadata={'units': 'degrees'})
    """A NumPy ndarray of shape (m,) representing the incidence angles in degrees."""

    def __repr__(self):
        """Custom representation showing only whether fields are set."""
        field_info = []
        for field_name, field_value in self.__dict__.items():
            if field_value is None:
                field_info.append(f"{field_name}: not set")
            else:
                field_info.append(f"{field_name}: set")
        return f"SurfaceEfficiency({', '.join(field_info)})"

    @iactsim_style
    def plot(self, heatmap: bool | None = None):
        """Plot the efficiency curves.

        Parameters
        ----------
        heatmap : bool, optional
            Whether to plot a 2D heat map.  If `None` (default), a heatmap
            will be used when the number of wavelengths and incidence angles
            is large (greater than 11 for the smaller of the two dimensions).
            Otherwise, individual curves will be plotted.

        Returns
        -------
        matplotlib.figure.Figure, matplotlib.axes.Axes
            The matplotlib figure and axes object containing the plot.
        """
        if (self.incidence_angle is None and self.wavelength is None) or self.value is None:
            raise(RuntimeError("The surface efficiency has not been initialized properly."))
        
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()

        max_n_labels = 11

        plot_wavelength = self.wavelength is not None
        plot_incidence_angle = self.incidence_angle is not None

        n_wl = len(self.wavelength) if plot_wavelength else 1
        n_th = len(self.incidence_angle) if plot_incidence_angle else 1
        values = self.value.flatten()

        if heatmap is None:
            heatmap = False
            if n_th > 1 and n_wl > 1:
                if min(n_th, n_wl) > max_n_labels:
                    heatmap = True

        if not plot_wavelength or plot_incidence_angle:
            heatmap = False

        if not heatmap:
            curves = values.reshape((n_th, n_wl))
            if n_th <= n_wl:
                for i in range(n_th):
                    label = None
                    if self.incidence_angle is not None:
                        label = f'{self.incidence_angle[i]:.1f} $\\degree$'
                    plt.plot(self.wavelength, curves[i], label=label)
                plt.xlabel('Wavelength (nm)')
            else:
                for i in range(n_wl):
                    label = None
                    if self.wavelength is not None:
                        label = f'{self.wavelength[i]:.1f} nm'
                    plt.plot(self.incidence_angle, curves[:,i], label=label)
                plt.xlabel('Incidence angle ($\\degree$)')
            
            if min(n_th, n_wl) < max_n_labels and (plot_wavelength or plot_incidence_angle):
                plt.legend()
            
            plt.ylabel('Efficiency')
            ax.grid(which='both')
        else:
            contour = ax.contourf(self.wavelength, self.incidence_angle, self.value, levels=np.linspace(0,1,100))
            ax.set_xlabel("Wavelength (nm)")
            ax.set_ylabel("Incidence angle ($\\degree$)")
            fig.colorbar(contour, ax=ax, boundaries=[0,1])
        
        return fig, ax


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import vtk

from ._vtk_utils import (
    create_aspheric_surface_actor,
    create_cylindric_surface_actor,
    quit
)

from ..optics._surface_misc import SurfaceType, SurfaceShape

class VTKOpticalSystem():
    """Class to viasualize the geometry of an optical system.
    Each surface actor can be accessed from :py:attr:`actors` attribute after a :py:meth:`update` call.

    Parameters
    ----------
    optical_system : OpticalSystem
        Optical sistem for which visualize surfaces.

    Notes
    -----
    If you perform some operation on an actor, make sure to call :py:meth:`start_render` with ``update=False``, otherwise the actors will be replaced.
    
    """
    def __init__(self, optical_system):
        self.actors = {}
        """Dictionary of surface name: surface actor."""

        self.os = optical_system
        """Optical system to be visualized."""
        
        self._default_surface_color = (0.7, 0.7, 0.75)

        self.surface_type_colors = {
            SurfaceType.REFRACTIVE: (0.6, 0.4, 1),
            SurfaceType.OPAQUE: (1, 0.3, 0.3)
        }
        """Dictionary to custumize surface colors based on surface type."""

        # Wireframe
        self.wireframe = False
        """Whether to use wireframe representation by default."""

        # Window size
        self.window_size = (1024,1024)
        """Window size in pixel."""

        self.resolution = 10. # mm
        """Objects mesh resolution (in mm)."""
    
    def update(self):
        """Generate all surface actors.
        """
        self.actors = {}
        for s in self.os:
            if s._shape == SurfaceShape.ASPHERICAL:
                
                actor = create_aspheric_surface_actor(s, self.resolution)
            else:
                actor = create_cylindric_surface_actor(s, self.resolution)

            transform = vtk.vtkTransform()
            transform.Translate(*s.position)
            R = s.get_rotation_matrix()
            vtk_rotation_matrix = vtk.vtkMatrix4x4()
            for i in range(3):
                for j in range(3):
                    vtk_rotation_matrix.SetElement(i, j, R[j, i])
            transform.PreMultiply()
            transform.Concatenate(vtk_rotation_matrix)
            transform.Update()
            actor.SetUserTransform(transform)
            
            if self.wireframe:
                actor.GetProperty().SetRepresentationToWireframe()
            
            if s.type in self.surface_type_colors:
                color = self.surface_type_colors[s.type]
            else:
                color = self._default_surface_color

            actor.GetProperty().SetColor(*color)
            actor.GetProperty().SetAmbient(0.3)
            actor.GetProperty().SetDiffuse(0.5)
            actor.GetProperty().SetSpecular(0.2)
            actor.GetProperty().SetSpecularPower(10)
        
            self.actors[s.name] = actor
    
    def start_render(self, update=True):
        """Render the optical system geometry on a VTK window.

        Parameters
        ----------
        update : bool, optional
            Regenerate surface actors, by default True.
            Make sure to not re-generate actors if you have modified them.
        
        """
        if update:
            self.update()

        # Rendering
        renderer = vtk.vtkRenderer()
        renderer.SetBackground(0.1, 0.15, 0.25)

        for actor in self.actors.values():
            renderer.AddActor(actor)

        # Display instructions
        desc = [
            "Press: 'q' to exit",
            "Press: 'r' to reset the camera",
            "Press: 'x', 'y' or 'z' to align the up-vector to the desired axis",
            "Press: 'w', 's' or 'p' to switch to wireframe, surface or points representation",
        ]
        textActor = vtk.vtkTextActor()
        textActor.SetInput('\n'.join(desc))
        position_coordinate = textActor.GetPositionCoordinate()
        position_coordinate.SetCoordinateSystemToNormalizedViewport()
        position_coordinate.SetValue(0.01, 0.99)
        textActor.GetTextProperty().SetJustificationToLeft()
        textActor.GetTextProperty().SetVerticalJustificationToTop()
        textActor.GetTextProperty().SetFontSize(14)
        textActor.GetTextProperty().SetColor(vtk.vtkNamedColors().GetColor3d("Gold"))
        textActor.GetTextProperty().SetFontFamily(vtk.VTK_COURIER)
        renderer.AddActor2D(textActor)

        render_window = vtk.vtkRenderWindow()
        render_window.SetWindowName(self.os.name)
        render_window.AddRenderer(renderer)
        render_window.SetSize(*self.window_size)
        render_window_interactor = vtk.vtkRenderWindowInteractor()
        render_window_interactor.SetRenderWindow(render_window)

        # Camera
        cam_style = vtk.vtkInteractorStyleTrackballCamera()
        render_window_interactor.SetInteractorStyle(cam_style)

        # Axes
        axes = vtk.vtkAxesActor()
        widget = vtk.vtkOrientationMarkerWidget()
        rgba = [0] * 4
        vtk.vtkNamedColors().GetColor('Carrot', rgba)
        widget.SetOutlineColor(rgba[0], rgba[1], rgba[2])
        widget.SetOrientationMarker(axes)
        widget.SetInteractor(render_window_interactor)
        widget.SetEnabled(1)
        widget.InteractiveOn()

        def perform_view_reset(up_vector):
            active_camera = renderer.GetActiveCamera()
            if active_camera:
                active_camera.SetViewUp(*up_vector)
                render_window.Render()

        def key_press_callback(caller, event):
            interactor = caller
            key_sym = interactor.GetKeySym()
            if key_sym and key_sym.lower() in 'r':
                renderer.ResetCamera()
            elif key_sym and key_sym.lower() == 'y':
                perform_view_reset((0,1,0))
            elif key_sym and key_sym.lower() == 'x':
                perform_view_reset((1,0,0))
            elif key_sym and key_sym.lower() == 'z':
                perform_view_reset((0,0,1))
            elif key_sym and key_sym.lower() == 'w':
                for actor in renderer.actors:
                    actor.GetProperty().SetRepresentationToWireframe()
                render_window.Render()
            elif key_sym and key_sym.lower() == 's':
                for actor in renderer.actors:
                    actor.GetProperty().SetRepresentationToSurface()
                render_window.Render()
            elif key_sym and key_sym.lower() == 'p':
                for actor in renderer.actors:
                    actor.GetProperty().SetRepresentationToPoints()
                render_window.Render()
            elif key_sym and key_sym.lower() == 'q':
                    quit(render_window_interactor)

        priority = 0.0
        render_window_interactor.AddObserver(vtk.vtkCommand.KeyPressEvent, key_press_callback, priority)

        # Start
        render_window.Render()
        renderer.ResetCamera()
        render_window.Render()
        render_window_interactor.Initialize()
        render_window_interactor.Start()

        # Stop
        quit(render_window_interactor)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import os as os
from pathlib import Path as Path
import cupy as cp
from ....utils._kernels import get_kernel

include_path = ''.join([os.environ['CPATH']])

path = Path(__file__).parent / "topological_trigger_logic.cu"
with open(path) as _source_file:
    source_code = _source_file.read()

module = cp.RawModule(
    code=source_code,
    backend='nvcc',
    options=(''.join(['-I',include_path]),'--use_fast_math', '-std=c++11', '--extra-device-vectorization', '-O3'),
    enable_cooperative_groups=True
)

topological_camera_trigger = get_kernel(module, 'topological_camera_trigger')
count_topological_camera_triggers = get_kernel(module, 'count_topological_camera_triggers')

# TODO: write an iterative implementation
# Increase the stack size of each GPU thread 
# (for contiguous pixels search recursion)
cp.cuda.runtime.deviceSetLimit(cp.cuda.runtime.cudaLimitStackSize, 16*1024) # bytes